@@ -4,7 +4,7 @@
  *                Alan C. Hindmarsh and Radu Serban @ LLNL
  * -----------------------------------------------------------------
  * SUNDIALS Copyright Start
- * Copyright (c) 2002-2019, Lawrence Livermore National Security
+ * Copyright (c) 2002-2022, Lawrence Livermore National Security
  * and Southern Methodist University.
  * All rights reserved.
  *
@@ -23,6 +23,9 @@
 #include <sunlinsol/sunlinsol_spgmr.h>
 #include <sundials/sundials_math.h>
 
+#include "sundials_context_impl.h"
+#include "sundials_logger_impl.h"
+
 #define ZERO RCONST(0.0)
 #define ONE  RCONST(1.0)
 
@@ -35,23 +38,6 @@
 #define SPGMR_CONTENT(S)  ( (SUNLinearSolverContent_SPGMR)(S->content) )
 #define LASTFLAG(S)       ( SPGMR_CONTENT(S)->last_flag )
 
-/*
- * -----------------------------------------------------------------
- * deprecated wrapper functions
- * -----------------------------------------------------------------
- */
-SUNLinearSolver SUNSPGMR(N_Vector y, int pretype, int maxl)
-{ return(SUNLinSol_SPGMR(y, pretype, maxl)); }
-
-int SUNSPGMRSetPrecType(SUNLinearSolver S, int pretype)
-{ return(SUNLinSol_SPGMRSetPrecType(S, pretype)); }
-
-int SUNSPGMRSetGSType(SUNLinearSolver S, int gstype)
-{ return(SUNLinSol_SPGMRSetGSType(S, gstype)); }
-
-int SUNSPGMRSetMaxRestarts(SUNLinearSolver S, int maxrs)
-{ return(SUNLinSol_SPGMRSetMaxRestarts(S, maxrs)); }
-
 /*
  * -----------------------------------------------------------------
  * exported functions
@@ -62,15 +48,15 @@ int SUNSPGMRSetMaxRestarts(SUNLinearSolver S, int maxrs)
  * Function to create a new SPGMR linear solver
  */
 
-SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl)
+SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl, SUNContext sunctx)
 {
   SUNLinearSolver S;
   SUNLinearSolverContent_SPGMR content;
 
   /* check for legal pretype and maxl values; if illegal use defaults */
-  if ((pretype != PREC_NONE)  && (pretype != PREC_LEFT) &&
-      (pretype != PREC_RIGHT) && (pretype != PREC_BOTH))
-    pretype = PREC_NONE;
+  if ((pretype != SUN_PREC_NONE)  && (pretype != SUN_PREC_LEFT) &&
+      (pretype != SUN_PREC_RIGHT) && (pretype != SUN_PREC_BOTH))
+    pretype = SUN_PREC_NONE;
   if (maxl <= 0)
     maxl = SUNSPGMR_MAXL_DEFAULT;
 
@@ -83,7 +69,7 @@ SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl)
 
   /* Create linear solver */
   S = NULL;
-  S = SUNLinSolNewEmpty();
+  S = SUNLinSolNewEmpty(sunctx);
   if (S == NULL) return(NULL);
 
   /* Attach operations */
@@ -92,6 +78,7 @@ SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl)
   S->ops->setatimes         = SUNLinSolSetATimes_SPGMR;
   S->ops->setpreconditioner = SUNLinSolSetPreconditioner_SPGMR;
   S->ops->setscalingvectors = SUNLinSolSetScalingVectors_SPGMR;
+  S->ops->setzeroguess      = SUNLinSolSetZeroGuess_SPGMR;
   S->ops->initialize        = SUNLinSolInitialize_SPGMR;
   S->ops->setup             = SUNLinSolSetup_SPGMR;
   S->ops->solve             = SUNLinSolSolve_SPGMR;
@@ -116,6 +103,7 @@ SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl)
   content->pretype      = pretype;
   content->gstype       = SUNSPGMR_GSTYPE_DEFAULT;
   content->max_restarts = SUNSPGMR_MAXRS_DEFAULT;
+  content->zeroguess    = SUNFALSE;
   content->numiters     = 0;
   content->resnorm      = ZERO;
   content->xcor         = NULL;
@@ -133,6 +121,11 @@ SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl)
   content->yg           = NULL;
   content->cv           = NULL;
   content->Xv           = NULL;
+  content->print_level  = 0;
+  content->info_file    = stdout;
+#if SUNDIALS_LOGGING_LEVEL >= SUNDIALS_LOGGING_INFO
+  content->info_file    = (sunctx->logger->info_fp) ? sunctx->logger->info_fp : stdout;
+#endif
 
   /* Allocate content */
   content->xcor = N_VClone(y);
@@ -149,11 +142,11 @@ SUNLinearSolver SUNLinSol_SPGMR(N_Vector y, int pretype, int maxl)
  * Function to set the type of preconditioning for SPGMR to use
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPGMRSetPrecType(SUNLinearSolver S, int pretype)
+int SUNLinSol_SPGMRSetPrecType(SUNLinearSolver S, int pretype)
 {
   /* Check for legal pretype */
-  if ((pretype != PREC_NONE)  && (pretype != PREC_LEFT) &&
-      (pretype != PREC_RIGHT) && (pretype != PREC_BOTH)) {
+  if ((pretype != SUN_PREC_NONE)  && (pretype != SUN_PREC_LEFT) &&
+      (pretype != SUN_PREC_RIGHT) && (pretype != SUN_PREC_BOTH)) {
     return(SUNLS_ILL_INPUT);
   }
 
@@ -170,10 +163,10 @@ SUNDIALS_EXPORT int SUNLinSol_SPGMRSetPrecType(SUNLinearSolver S, int pretype)
  * Function to set the type of Gram-Schmidt orthogonalization for SPGMR to use
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPGMRSetGSType(SUNLinearSolver S, int gstype)
+int SUNLinSol_SPGMRSetGSType(SUNLinearSolver S, int gstype)
 {
   /* Check for legal gstype */
-  if ((gstype != MODIFIED_GS) && (gstype != CLASSICAL_GS)) {
+  if ((gstype != SUN_MODIFIED_GS) && (gstype != SUN_CLASSICAL_GS)) {
     return(SUNLS_ILL_INPUT);
   }
 
@@ -190,7 +183,7 @@ SUNDIALS_EXPORT int SUNLinSol_SPGMRSetGSType(SUNLinearSolver S, int gstype)
  * Function to set the maximum number of GMRES restarts to allow
  */
 
-SUNDIALS_EXPORT int SUNLinSol_SPGMRSetMaxRestarts(SUNLinearSolver S, int maxrs)
+int SUNLinSol_SPGMRSetMaxRestarts(SUNLinearSolver S, int maxrs)
 {
   /* Illegal maxrs implies use of default value */
   if (maxrs < 0)
@@ -235,11 +228,21 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   /* ensure valid options */
   if (content->max_restarts < 0)
     content->max_restarts = SUNSPGMR_MAXRS_DEFAULT;
-  if ( (content->pretype != PREC_LEFT) &&
-       (content->pretype != PREC_RIGHT) &&
-       (content->pretype != PREC_BOTH) )
-    content->pretype = PREC_NONE;
 
+  if (content->ATimes == NULL) {
+    LASTFLAG(S) = SUNLS_ATIMES_NULL;
+    return(LASTFLAG(S));
+  }
+
+  if ( (content->pretype != SUN_PREC_LEFT) &&
+       (content->pretype != SUN_PREC_RIGHT) &&
+       (content->pretype != SUN_PREC_BOTH) )
+    content->pretype = SUN_PREC_NONE;
+
+  if ((content->pretype != SUN_PREC_NONE) && (content->Psolve == NULL)) {
+    LASTFLAG(S) = SUNLS_PSOLVE_NULL;
+    return(LASTFLAG(S));
+  }
 
   /* allocate solver-specific memory (where the size depends on the
      choice of maxl) here */
@@ -248,7 +251,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   if (content->V == NULL) {
     content->V = N_VCloneVectorArray(content->maxl+1, content->vtemp);
     if (content->V == NULL) {
-      SUNLinSolFree(S);
       content->last_flag = SUNLS_MEM_FAIL;
       return(SUNLS_MEM_FAIL);
     }
@@ -258,7 +260,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   if (content->Hes == NULL) {
     content->Hes = (realtype **) malloc((content->maxl+1)*sizeof(realtype *));
     if (content->Hes == NULL) {
-      SUNLinSolFree(S);
       content->last_flag = SUNLS_MEM_FAIL;
       return(SUNLS_MEM_FAIL);
     }
@@ -267,7 +268,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
       content->Hes[k] = NULL;
       content->Hes[k] = (realtype *) malloc(content->maxl*sizeof(realtype));
       if (content->Hes[k] == NULL) {
-        SUNLinSolFree(S);
         content->last_flag = SUNLS_MEM_FAIL;
         return(SUNLS_MEM_FAIL);
       }
@@ -278,7 +278,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   if (content->givens == NULL) {
     content->givens = (realtype *) malloc(2*content->maxl*sizeof(realtype));
     if (content->givens == NULL) {
-      SUNLinSolFree(S);
       content->last_flag = SUNLS_MEM_FAIL;
       return(SUNLS_MEM_FAIL);
     }
@@ -288,7 +287,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   if (content->yg == NULL) {
     content->yg = (realtype *) malloc((content->maxl+1)*sizeof(realtype));
     if (content->yg == NULL) {
-      SUNLinSolFree(S);
       content->last_flag = SUNLS_MEM_FAIL;
       return(SUNLS_MEM_FAIL);
     }
@@ -298,7 +296,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   if (content->cv == NULL) {
     content->cv = (realtype *) malloc((content->maxl+1)*sizeof(realtype));
     if (content->cv == NULL) {
-      SUNLinSolFree(S);
       content->last_flag = SUNLS_MEM_FAIL;
       return(SUNLS_MEM_FAIL);
     }
@@ -308,7 +305,6 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
   if (content->Xv == NULL) {
     content->Xv = (N_Vector *) malloc((content->maxl+1)*sizeof(N_Vector));
     if (content->Xv == NULL) {
-      SUNLinSolFree(S);
       content->last_flag = SUNLS_MEM_FAIL;
       return(SUNLS_MEM_FAIL);
     }
@@ -321,7 +317,7 @@ int SUNLinSolInitialize_SPGMR(SUNLinearSolver S)
 
 
 int SUNLinSolSetATimes_SPGMR(SUNLinearSolver S, void* ATData,
-                             ATimesFn ATimes)
+                             SUNATimesFn ATimes)
 {
   /* set function pointers to integrator-supplied ATimes routine
      and data, and return with success */
@@ -334,7 +330,7 @@ int SUNLinSolSetATimes_SPGMR(SUNLinearSolver S, void* ATData,
 
 
 int SUNLinSolSetPreconditioner_SPGMR(SUNLinearSolver S, void* PData,
-                                    PSetupFn Psetup, PSolveFn Psolve)
+                                     SUNPSetupFn Psetup, SUNPSolveFn Psolve)
 {
   /* set function pointers to integrator-supplied Psetup and PSolve
      routines and data, and return with success */
@@ -360,10 +356,20 @@ int SUNLinSolSetScalingVectors_SPGMR(SUNLinearSolver S, N_Vector s1,
 }
 
 
+int SUNLinSolSetZeroGuess_SPGMR(SUNLinearSolver S, booleantype onff)
+{
+  /* set flag indicating a zero initial guess */
+  if (S == NULL) return(SUNLS_MEM_NULL);
+  SPGMR_CONTENT(S)->zeroguess = onff;
+  LASTFLAG(S) = SUNLS_SUCCESS;
+  return(LASTFLAG(S));
+}
+
+
 int SUNLinSolSetup_SPGMR(SUNLinearSolver S, SUNMatrix A)
 {
   int ier;
-  PSetupFn Psetup;
+  SUNPSetupFn Psetup;
   void* PData;
 
   /* Set shortcuts to SPGMR memory structures */
@@ -377,7 +383,7 @@ int SUNLinSolSetup_SPGMR(SUNLinearSolver S, SUNMatrix A)
     ier = Psetup(PData);
     if (ier != 0) {
       LASTFLAG(S) = (ier < 0) ?
-	SUNLS_PSET_FAIL_UNREC : SUNLS_PSET_FAIL_REC;
+        SUNLS_PSET_FAIL_UNREC : SUNLS_PSET_FAIL_REC;
       return(LASTFLAG(S));
     }
   }
@@ -395,11 +401,12 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   realtype **Hes, *givens, *yg, *res_norm;
   realtype beta, rotation_product, r_norm, s_product, rho;
   booleantype preOnLeft, preOnRight, scale2, scale1, converged;
+  booleantype *zeroguess;
   int i, j, k, l, l_plus_1, l_max, krydim, ier, ntries, max_restarts, gstype;
   int *nli;
   void *A_data, *P_data;
-  ATimesFn atimes;
-  PSolveFn psolve;
+  SUNATimesFn atimes;
+  SUNPSolveFn psolve;
 
   /* local shortcuts for fused vector operations */
   realtype* cv;
@@ -426,6 +433,7 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   P_data       = SPGMR_CONTENT(S)->PData;
   atimes       = SPGMR_CONTENT(S)->ATimes;
   psolve       = SPGMR_CONTENT(S)->Psolve;
+  zeroguess    = &(SPGMR_CONTENT(S)->zeroguess);
   nli          = &(SPGMR_CONTENT(S)->numiters);
   res_norm     = &(SPGMR_CONTENT(S)->resnorm);
   cv           = SPGMR_CONTENT(S)->cv;
@@ -435,22 +443,43 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   *nli = 0;
   converged = SUNFALSE;
 
-  /* set booleantype flags for internal solver options */
-  preOnLeft  = ( (SPGMR_CONTENT(S)->pretype == PREC_LEFT) ||
-                 (SPGMR_CONTENT(S)->pretype == PREC_BOTH) );
-  preOnRight = ( (SPGMR_CONTENT(S)->pretype == PREC_RIGHT) ||
-                 (SPGMR_CONTENT(S)->pretype == PREC_BOTH) );
+  /* Set booleantype flags for internal solver options */
+  preOnLeft  = ( (SPGMR_CONTENT(S)->pretype == SUN_PREC_LEFT) ||
+                 (SPGMR_CONTENT(S)->pretype == SUN_PREC_BOTH) );
+  preOnRight = ( (SPGMR_CONTENT(S)->pretype == SUN_PREC_RIGHT) ||
+                 (SPGMR_CONTENT(S)->pretype == SUN_PREC_BOTH) );
   scale1 = (s1 != NULL);
   scale2 = (s2 != NULL);
 
+#if SUNDIALS_LOGGING_LEVEL >= SUNDIALS_LOGGING_INFO
+  if (SPGMR_CONTENT(S)->print_level && SPGMR_CONTENT(S)->info_file
+      && (SPGMR_CONTENT(S)->info_file != S->sunctx->logger->info_fp))
+    fprintf(SPGMR_CONTENT(S)->info_file, "SUNLINSOL_SPGMR:\n");
+#endif
+
+  /* Check if Atimes function has been set */
+  if (atimes == NULL) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_ATIMES_NULL;
+    return(LASTFLAG(S));
+  }
+
+  /* If preconditioning, check if psolve has been set */
+  if ((preOnLeft || preOnRight) && psolve == NULL) {
+    *zeroguess  = SUNFALSE;
+    LASTFLAG(S) = SUNLS_PSOLVE_NULL;
+    return(LASTFLAG(S));
+  }
+
   /* Set vtemp and V[0] to initial (unscaled) residual r_0 = b - A*x_0 */
-  if (N_VDotProd(x, x) == ZERO) {
+  if (*zeroguess) {
     N_VScale(ONE, b, vtemp);
   } else {
     ier = atimes(A_data, x, vtemp);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
-          SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
+        SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
       return(LASTFLAG(S));
     }
     N_VLinearSum(ONE, b, -ONE, vtemp, vtemp);
@@ -459,10 +488,11 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
 
   /* Apply left preconditioner and left scaling to V[0] = r_0 */
   if (preOnLeft) {
-    ier = psolve(P_data, V[0], vtemp, delta, PREC_LEFT);
+    ier = psolve(P_data, V[0], vtemp, delta, SUN_PREC_LEFT);
     if (ier != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = (ier < 0) ?
-          SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
+        SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
       return(LASTFLAG(S));
     }
   } else {
@@ -478,7 +508,23 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
   /* Set r_norm = beta to L2 norm of V[0] = s1 P1_inv r_0, and
      return if small  */
   *res_norm = r_norm = beta = SUNRsqrt(N_VDotProd(V[0], V[0]));
+
+#if SUNDIALS_LOGGING_LEVEL >= SUNDIALS_LOGGING_INFO
+  /* print initial residual */
+  if (SPGMR_CONTENT(S)->print_level && SPGMR_CONTENT(S)->info_file
+      && (SPGMR_CONTENT(S)->info_file != S->sunctx->logger->info_fp))
+  {
+    fprintf(SPGMR_CONTENT(S)->info_file,
+            SUNLS_MSG_RESIDUAL,
+            (long int) 0, *res_norm);
+  }
+  SUNLogger_QueueMsg(S->sunctx->logger, SUN_LOGLEVEL_INFO,
+    "SUNLinSolSolve_SPGMR", "initial-residual",
+    "nli = %li, resnorm = %.16g", (long int) 0, *res_norm);
+#endif
+
   if (r_norm <= delta) {
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = SUNLS_SUCCESS;
     return(LASTFLAG(S));
   }
@@ -515,10 +561,11 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       /*   Apply right preconditioner: vtemp = P2_inv s2_inv V[l] */
       if (preOnRight) {
         N_VScale(ONE, vtemp, V[l_plus_1]);
-        ier = psolve(P_data, V[l_plus_1], vtemp, delta, PREC_RIGHT);
+        ier = psolve(P_data, V[l_plus_1], vtemp, delta, SUN_PREC_RIGHT);
         if (ier != 0) {
+          *zeroguess  = SUNFALSE;
           LASTFLAG(S) = (ier < 0) ?
-              SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
+            SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
           return(LASTFLAG(S));
         }
       }
@@ -526,17 +573,19 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       /* Apply A: V[l+1] = A P2_inv s2_inv V[l] */
       ier = atimes( A_data, vtemp, V[l_plus_1] );
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
-            SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
+          SUNLS_ATIMES_FAIL_UNREC : SUNLS_ATIMES_FAIL_REC;
         return(LASTFLAG(S));
       }
 
       /* Apply left preconditioning: vtemp = P1_inv A P2_inv s2_inv V[l] */
       if (preOnLeft) {
-        ier = psolve(P_data, V[l_plus_1], vtemp, delta, PREC_LEFT);
+        ier = psolve(P_data, V[l_plus_1], vtemp, delta, SUN_PREC_LEFT);
         if (ier != 0) {
+          *zeroguess  = SUNFALSE;
           LASTFLAG(S) = (ier < 0) ?
-              SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
+            SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
           return(LASTFLAG(S));
         }
       } else {
@@ -551,21 +600,24 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       }
 
       /*  Orthogonalize V[l+1] against previous V[i]: V[l+1] = w_tilde */
-      if (gstype == CLASSICAL_GS) {
-        if (ClassicalGS(V, Hes, l_plus_1, l_max, &(Hes[l_plus_1][l]),
-                        cv, Xv) != 0) {
+      if (gstype == SUN_CLASSICAL_GS) {
+        if (SUNClassicalGS(V, Hes, l_plus_1, l_max, &(Hes[l_plus_1][l]),
+                           cv, Xv) != 0) {
+          *zeroguess  = SUNFALSE;
           LASTFLAG(S) = SUNLS_GS_FAIL;
           return(LASTFLAG(S));
         }
       } else {
-        if (ModifiedGS(V, Hes, l_plus_1, l_max, &(Hes[l_plus_1][l])) != 0) {
+        if (SUNModifiedGS(V, Hes, l_plus_1, l_max, &(Hes[l_plus_1][l])) != 0) {
+          *zeroguess  = SUNFALSE;
           LASTFLAG(S) = SUNLS_GS_FAIL;
           return(LASTFLAG(S));
         }
       }
 
       /*  Update the QR factorization of Hes */
-      if(QRfact(krydim, Hes, givens, l) != 0 ) {
+      if(SUNQRfact(krydim, Hes, givens, l) != 0 ) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = SUNLS_QRFACT_FAIL;
         return(LASTFLAG(S));
       }
@@ -574,6 +626,20 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       rotation_product *= givens[2*l+1];
       *res_norm = rho = SUNRabs(rotation_product*r_norm);
 
+#if SUNDIALS_LOGGING_LEVEL >= SUNDIALS_LOGGING_INFO
+      /* print current iteration number and the residual */
+      if (SPGMR_CONTENT(S)->print_level && SPGMR_CONTENT(S)->info_file
+          && (SPGMR_CONTENT(S)->info_file != S->sunctx->logger->info_fp))
+      {
+        fprintf(SPGMR_CONTENT(S)->info_file,
+                SUNLS_MSG_RESIDUAL,
+                (long int) *nli, *res_norm);
+      }
+      SUNLogger_QueueMsg(S->sunctx->logger, SUN_LOGLEVEL_INFO,
+        "SUNLinSolSolve_SPGMR", "iterate-residual",
+        "nli = %li, resnorm = %.16g", (long int) *nli, *res_norm);
+#endif
+
       if (rho <= delta) { converged = SUNTRUE; break; }
 
       /* Normalize V[l+1] with norm value from the Gram-Schmidt routine */
@@ -585,7 +651,8 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     /*   Construct g, then solve for y */
     yg[0] = r_norm;
     for (i=1; i<=krydim; i++) yg[i]=ZERO;
-    if (QRsol(krydim, Hes, givens, yg) != 0) {
+    if (SUNQRsol(krydim, Hes, givens, yg) != 0) {
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = SUNLS_QRSOL_FAIL;
       return(LASTFLAG(S));
     }
@@ -599,7 +666,11 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       Xv[k+1] = V[k];
     }
     ier = N_VLinearCombination(krydim+1, cv, Xv, xcor);
-    if (ier != SUNLS_SUCCESS) return(SUNLS_VECTOROP_ERR);
+    if (ier != SUNLS_SUCCESS) {
+      *zeroguess  = SUNFALSE;
+      LASTFLAG(S) = SUNLS_VECTOROP_ERR;
+      return(SUNLS_VECTOROP_ERR);
+    }
 
     /* If converged, construct the final solution vector x and return */
     if (converged) {
@@ -607,10 +678,11 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       /* Apply right scaling and right precond.: vtemp = P2_inv s2_inv xcor */
       if (scale2) N_VDiv(xcor, s2, xcor);
       if (preOnRight) {
-        ier = psolve(P_data, xcor, vtemp, delta, PREC_RIGHT);
+        ier = psolve(P_data, xcor, vtemp, delta, SUN_PREC_RIGHT);
         if (ier != 0) {
+          *zeroguess  = SUNFALSE;
           LASTFLAG(S) = (ier < 0) ?
-              SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
+            SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
           return(LASTFLAG(S));
         }
       } else {
@@ -618,8 +690,12 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       }
 
       /* Add vtemp to initial x to get final solution x, and return */
-      N_VLinearSum(ONE, x, ONE, vtemp, x);
+      if (*zeroguess)
+        N_VScale(ONE, vtemp, x);
+      else
+        N_VLinearSum(ONE, x, ONE, vtemp, x);
 
+      *zeroguess  = SUNFALSE;
       LASTFLAG(S) = SUNLS_SUCCESS;
       return(LASTFLAG(S));
     }
@@ -647,7 +723,11 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
       Xv[k] = V[k];
     }
     ier = N_VLinearCombination(krydim+1, cv, Xv, V[0]);
-    if (ier != SUNLS_SUCCESS) return(SUNLS_VECTOROP_ERR);
+    if (ier != SUNLS_SUCCESS) {
+      *zeroguess  = SUNFALSE;
+      LASTFLAG(S) = SUNLS_VECTOROP_ERR;
+      return(SUNLS_VECTOROP_ERR);
+    }
 
   }
 
@@ -659,23 +739,29 @@ int SUNLinSolSolve_SPGMR(SUNLinearSolver S, SUNMatrix A, N_Vector x,
     /* Apply right scaling and right precond.: vtemp = P2_inv s2_inv xcor */
     if (scale2) N_VDiv(xcor, s2, xcor);
     if (preOnRight) {
-      ier = psolve(P_data, xcor, vtemp, delta, PREC_RIGHT);
+      ier = psolve(P_data, xcor, vtemp, delta, SUN_PREC_RIGHT);
       if (ier != 0) {
+        *zeroguess  = SUNFALSE;
         LASTFLAG(S) = (ier < 0) ?
-            SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
+          SUNLS_PSOLVE_FAIL_UNREC : SUNLS_PSOLVE_FAIL_REC;
         return(LASTFLAG(S));
       }
-      } else {
+    } else {
       N_VScale(ONE, xcor, vtemp);
     }
 
     /* Add vtemp to initial x to get final solution x, and return */
-    N_VLinearSum(ONE, x, ONE, vtemp, x);
+    if (*zeroguess)
+      N_VScale(ONE, vtemp, x);
+    else
+      N_VLinearSum(ONE, x, ONE, vtemp, x);
 
+    *zeroguess  = SUNFALSE;
     LASTFLAG(S) = SUNLS_RES_REDUCED;
     return(LASTFLAG(S));
   }
 
+  *zeroguess  = SUNFALSE;
   LASTFLAG(S) = SUNLS_CONV_FAIL;
   return(LASTFLAG(S));
 }
@@ -781,3 +867,33 @@ int SUNLinSolFree_SPGMR(SUNLinearSolver S)
   free(S); S = NULL;
   return(SUNLS_SUCCESS);
 }
+
+
+int SUNLinSolSetInfoFile_SPGMR(SUNLinearSolver S,
+                               FILE* info_file)
+{
+  /* check that the linear solver is non-null */
+  if (S == NULL)
+    return(SUNLS_MEM_NULL);
+
+  SPGMR_CONTENT(S)->info_file = info_file;
+
+  return(SUNLS_SUCCESS);
+}
+
+
+int SUNLinSolSetPrintLevel_SPGMR(SUNLinearSolver S,
+                                 int print_level)
+{
+  /* check that the linear solver is non-null */
+  if (S == NULL)
+    return(SUNLS_MEM_NULL);
+
+  /* check for valid print level */
+  if (print_level < 0 || print_level > 1)
+    return(SUNLS_ILL_INPUT);
+
+  SPGMR_CONTENT(S)->print_level = print_level;
+
+  return(SUNLS_SUCCESS);
+}
